import os
from pathlib import Path
from typing import Optional

from tqdm import tqdm

from supervisely._utils import is_production
from supervisely.api.api import Api
from supervisely.app import get_data_dir
from supervisely.convert.image.csv.csv_converter import CSVConverter
from supervisely.convert.image.high_color.high_color_depth import (
    HighColorDepthImageConverter,
)
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.convert.pointcloud_episodes.pointcloud_episodes_converter import (
    PointcloudEpisodeConverter,
)
from supervisely.convert.video.video_converter import VideoConverter
from supervisely.convert.volume.volume_converter import VolumeConverter
from supervisely.io.env import team_id as env_team_id
from supervisely.io.fs import (
    dir_exists,
    file_exists,
    is_archive,
    mkdir,
    remove_junk_from_dir,
    silent_remove,
    touch,
    unpack_archive,
)
from supervisely.project.project_settings import LabelingInterface
from supervisely.project.project_type import ProjectType
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class ImportManager:

    def __init__(
        self,
        input_data: str,
        project_type: ProjectType,
        team_id: Optional[int] = None,
        labeling_interface: LabelingInterface = LabelingInterface.DEFAULT,
        upload_as_links: bool = False,
    ):
        self._api = Api.from_env()
        if team_id is not None:
            team_info = self._api.team.get_info_by_id(team_id)
            if team_info is None:
                raise ValueError(
                    f"Team with id {team_id} does not exist or you do not have access to it."
                )
        else:
            self._team_id = env_team_id()
        self._labeling_interface = labeling_interface
        self._upload_as_links = upload_as_links
        self._remote_files_map = {}

        self._input_data = self._prepare_input_data(input_data)
        self._unpack_archives(self._input_data)
        remove_junk_from_dir(self._input_data)

        self._modality = project_type
        self._converter = self.get_converter()
        if isinstance(self._converter, (HighColorDepthImageConverter, CSVConverter)):
            self._converter.team_id = self._team_id

    @property
    def modality(self):
        return self._modality

    @property
    def converter(self):
        return self._converter

    def get_items(self):
        return self._converter.get_items()

    def get_converter(self):
        """Return correct converter"""
        modality_converter_map = {
            ProjectType.IMAGES.value: ImageConverter,
            ProjectType.VIDEOS.value: VideoConverter,
            ProjectType.POINT_CLOUDS.value: PointcloudConverter,
            ProjectType.VOLUMES.value: VolumeConverter,
            ProjectType.POINT_CLOUD_EPISODES.value: PointcloudEpisodeConverter,
        }
        if str(self._modality) not in modality_converter_map:
            raise ValueError(f"Unsupported project type selected: {self._modality}")
        modality_converter = modality_converter_map[str(self._modality)](
            self._input_data,
            self._labeling_interface,
            self._upload_as_links,
            self._remote_files_map,
        )
        return modality_converter.detect_format()

    def upload_dataset(self, dataset_id):
        """Upload converted data to Supervisely"""
        self.converter.upload_dataset(self._api, dataset_id)

    # def validate_format(self):
    #     raise NotImplementedError

    def _prepare_input_data(self, input_data):
        if dir_exists(input_data):
            logger.info(f"Input data is a local directory: {input_data}")
            return input_data
        elif file_exists(input_data):
            logger.info(f"Input data is a local file: {input_data}. Will use its directory")
            return os.path.dirname(input_data)
        elif self._api.storage.exists(self._team_id, input_data):
            if self._upload_as_links:
                logger.info(f"Input data is a remote file: {input_data}. Scanning...")
                return self._scan_remote_files(input_data)
            else:
                logger.info(f"Input data is a remote file: {input_data}. Downloading...")
                return self._download_input_data(input_data)
        elif self._api.storage.dir_exists(self._team_id, input_data):
            if self._upload_as_links:
                logger.info(f"Input data is a remote directory: {input_data}. Scanning...")
                return self._scan_remote_files(input_data, is_dir=True)
            else:
                logger.info(f"Input data is a remote directory: {input_data}. Downloading...")
                return self._download_input_data(input_data, is_dir=True)
        else:
            raise RuntimeError(f"Input data not found: {input_data}")

    def _download_input_data(self, remote_path, is_dir=False):
        """Download input data from Supervisely"""

        if not is_dir:
            dir_name = "Import data"
            local_path = os.path.join(get_data_dir(), dir_name)
            mkdir(local_path, remove_content_if_exists=True)
            save_path = os.path.join(local_path, os.path.basename(remote_path))
        else:
            dir_name = os.path.basename(os.path.normpath(remote_path))
            local_path = os.path.join(get_data_dir(), dir_name)

        if not is_dir:
            files_size = self._api.storage.get_info_by_path(self._team_id, remote_path).sizeb
            progress, progress_cb = self._get_progress(files_size)
            self._api.storage.download(
                self._team_id, remote_path, save_path, progress_cb=progress_cb
            )
            if not is_production():
                progress.close()
        else:
            directory_size = self._api.storage.get_directory_size(self._team_id, remote_path)
            progress, progress_cb = self._get_progress(directory_size)
            self._api.storage.download_directory(
                self._team_id, remote_path, local_path, progress_cb=progress_cb
            )
            if not is_production():
                progress.close()

        return local_path

    def _scan_remote_files(self, remote_path, is_dir=False):
        """Scan remote directory and create dummy structure locally"""

        dir_path = remote_path.rstrip("/") if is_dir else os.path.dirname(remote_path)
        dir_name = os.path.basename(dir_path)

        local_path = os.path.abspath(os.path.join(get_data_dir(), dir_name))
        mkdir(local_path, remove_content_if_exists=True)

        if is_dir:
            files = self._api.storage.list(self._team_id, remote_path)
        else:
            files = [self._api.storage.get_info_by_path(self._team_id, remote_path)]

        for file in files:
            new_path = file.path.replace(dir_path, local_path)
            self._remote_files_map[new_path] = file.path
            Path(new_path).parent.mkdir(parents=True, exist_ok=True)
            touch(new_path)

        return local_path

    def _unpack_archives(self, local_path):
        """Unpack if input data contains an archive."""

        new_paths_to_scan = [local_path]
        while len(new_paths_to_scan) > 0:
            archives = []
            path = new_paths_to_scan.pop()
            for root, _, files in os.walk(path):
                for file in files:
                    file_path = os.path.join(root, file)
                    if is_archive(file_path=file_path):
                        try:
                            new_path = file_path.replace("".join(Path(file_path).suffixes), "")
                            unpack_archive(file_path, new_path)
                            archives.append(file_path)
                            new_paths_to_scan.append(new_path)
                        except Exception as e:
                            logger.error(f"Error while unpacking '{file}': {repr(e)}")

            for archive in archives:
                silent_remove(archive)

    def _get_progress(
        self,
        total: int,
        message: str = "Downloading...",
        is_size: bool = True,
    ) -> tuple:
        if is_production():
            progress = Progress(message, total, is_size=is_size)
            progress_cb = progress.iters_done_report
        else:
            progress = tqdm(
                total=total, desc=message, unit="B" if is_size else "it", unit_scale=is_size
            )
            progress_cb = progress.update
        return progress, progress_cb



import os
import pickle
from typing import List, Optional, Union

import yaml

from supervisely.app.widgets import SlyTqdm
from supervisely.task.progress import tqdm_sly


class BaseEvaluator:
    EVALUATION_PARAMS_YAML_PATH: str = None

    def __init__(
        self,
        gt_project_path: str,
        dt_project_path: str,
        result_dir: str = "./evaluation",
        progress: Optional[SlyTqdm] = None,
        items_count: Optional[int] = None,  # TODO: is it needed?
        classes_whitelist: Optional[List[str]] = None,
        evaluation_params: Optional[dict] = None,
    ):
        self.gt_project_path = gt_project_path
        self.dt_project_path = dt_project_path
        self.result_dir = result_dir
        self.total_items = items_count
        self.pbar = progress or tqdm_sly
        os.makedirs(result_dir, exist_ok=True)
        self.classes_whitelist = classes_whitelist

        if evaluation_params is None:
            evaluation_params = self._get_default_evaluation_params()
        self.evaluation_params = evaluation_params
        if self.evaluation_params:
            self.validate_evaluation_params(self.evaluation_params)

    def evaluate(self):
        raise NotImplementedError()

    def get_result_dir(self) -> str:
        return self.result_dir

    @classmethod
    def load_yaml_evaluation_params(cls) -> Union[str, None]:
        if cls.EVALUATION_PARAMS_YAML_PATH is None:
            return None
        with open(cls.EVALUATION_PARAMS_YAML_PATH, "r") as f:
            return f.read()

    @classmethod
    def validate_evaluation_params(cls, evaluation_params: dict) -> None:
        pass

    @classmethod
    def _get_default_evaluation_params(cls) -> dict:
        if cls.EVALUATION_PARAMS_YAML_PATH is None:
            return {}
        else:
            return yaml.safe_load(cls.load_yaml_evaluation_params())

    def _dump_pickle(self, data, file_path):
        with open(file_path, "wb") as f:
            pickle.dump(data, f)



import os

from supervisely.io.json import dump_json_file
from supervisely.nn.benchmark.coco_utils import read_coco_datasets, sly2coco
from supervisely.nn.benchmark.evaluation import BaseEvaluator
from supervisely.nn.benchmark.evaluation.coco import calculate_metrics
from pathlib import Path


class ObjectDetectionEvaluator(BaseEvaluator):
    EVALUATION_PARAMS_YAML_PATH = f"{Path(__file__).parent}/coco/evaluation_params.yaml"

    def evaluate(self):
        try:
            self.cocoGt_json, self.cocoDt_json = self._convert_to_coco()
        except AssertionError as e:
            raise ValueError(
                f"{e}. Please make sure that your GT and DT projects are correct. "
                "If GT project has nested datasets and DT project was crated with NN app, "
                "try to use newer version of NN app."
            )
        self.cocoGt, self.cocoDt = read_coco_datasets(self.cocoGt_json, self.cocoDt_json)
        with self.pbar(message="Evaluation: Calculating metrics", total=5) as p:
            self.eval_data = calculate_metrics(
                self.cocoGt,
                self.cocoDt,
                iouType="bbox",
                progress_cb=p.update,
                evaluation_params=self.evaluation_params,
            )
        self._dump_eval_results()

    @classmethod
    def validate_evaluation_params(cls, evaluation_params: dict) -> None:
        iou_threshold = evaluation_params.get("iou_threshold")
        if iou_threshold is not None:
            assert iou_threshold in [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95], (
                f"iou_threshold must be one of [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95], "
                f"but got {iou_threshold}"
            )

    def _convert_to_coco(self):
        cocoGt_json = sly2coco(
            self.gt_project_path,
            is_dt_dataset=False,
            accepted_shapes=["rectangle"],
            progress=self.pbar,
            classes_whitelist=self.classes_whitelist,
        )
        cocoDt_json = sly2coco(
            self.dt_project_path,
            is_dt_dataset=True,
            accepted_shapes=["rectangle"],
            progress=self.pbar,
            classes_whitelist=self.classes_whitelist,
        )

        if len(cocoGt_json["annotations"]) == 0:
            raise ValueError("Not found any annotations in GT project")
        if len(cocoDt_json["annotations"]) == 0:
            raise ValueError(
                "Not found any predictions. "
                "Please make sure that your model produces predictions."
            )
        assert (
            cocoDt_json["categories"] == cocoGt_json["categories"]
        ), "Classes in GT and Pred projects must be the same"
        assert [f'{x["dataset"]}/{x["file_name"]}' for x in cocoDt_json["images"]] == [
            f'{x["dataset"]}/{x["file_name"]}' for x in cocoGt_json["images"]
        ], "Images in GT and DT projects are different"
        return cocoGt_json, cocoDt_json

    def _dump_eval_results(self):
        cocoGt_path, cocoDt_path, eval_data_path = self._get_eval_paths()
        dump_json_file(self.cocoGt_json, cocoGt_path, indent=None)
        dump_json_file(self.cocoDt_json, cocoDt_path, indent=None)
        self._dump_pickle(self.eval_data, eval_data_path)

    def _get_eval_paths(self):
        base_dir = self.result_dir
        cocoGt_path = os.path.join(base_dir, "cocoGt.json")
        cocoDt_path = os.path.join(base_dir, "cocoDt.json")
        eval_data_path = os.path.join(base_dir, "eval_data.pkl")
        return cocoGt_path, cocoDt_path, eval_data_path



# vim: expandtab:ts=4:sw=4
import numpy as np


class Detection(object):
    """
    This class represents a bounding box detection in a single image.

    Parameters
    ----------
    tlwh : array_like
        Bounding box in format `(x, y, w, h)`.
    confidence : float
        Detector confidence score.
    feature : array_like
        A feature vector that describes the object contained in this image.

    Attributes
    ----------
    tlwh : ndarray
        Bounding box in format `(top left x, top left y, width, height)`.
    confidence : ndarray
        Detector confidence score.
    feature : ndarray | NoneType
        A feature vector that describes the object contained in this image.

    """

    def __init__(self, tlwh, confidence, feature):
        self.tlwh = np.asarray(tlwh, dtype=float)
        self.confidence = float(confidence)
        self.feature = np.asarray(feature, dtype=np.float32)

    def to_tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    def to_xyah(self):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = self.tlwh.copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret



from supervisely.nn.benchmark.evaluation.base_evaluator import BaseEvaluator
from supervisely.nn.benchmark.evaluation.object_detection_evaluator import ObjectDetectionEvaluator
from supervisely.nn.benchmark.evaluation.instance_segmentation_evaluator import InstanceSegmentationEvaluator



markdown_speedtest_intro = """## Inference Speed

This is a speed test benchmark for this model. The model was tested with the following configuration:

- **Device**: {}
- **Hardware**: {}
- **Runtime**: {}
"""

markdown_speedtest_overview = """
The table below shows the speed test results. For each test, the time taken to process one batch of images is shown, as well as the model's throughput (i.e, the number of images processed per second, or FPS). Results are averaged across **{}** iterations.
"""

markdown_real_time_inference = """## Real-time Inference

This chart compares different runtimes and devices (CPU or GPU)."""

# We additionally divide **predict** procedure into three stages: pre-process, inference, and post-process. Each bar in this chart consists of these three stages. For example, in the chart you can find how long the post-process phase lasts in a CPU device with an ONNXRuntime environment."""


markdown_batch_inference = """
This chart shows how the model's speed changes with different batch sizes . As the batch size increases, you can observe an increase in FPS (images per second).
"""



from typing import List


def generate_main_template(metric_visualizations: List):
    template_str = """<div class="model-benchmark-body">

    <sly-style>
    .model-benchmark-body .sly-markdown-widget .markdown-body { padding: 0; font-family: inherit; }
    .model-benchmark-body .sly-markdown-widget .markdown-body h2 { font-size: 18px; font-weight: 600; margin-bottom: 0px; border: 0; }
    .model-benchmark-body .sly-markdown-widget .markdown-body h3 { color: #949bab; font-size: 18px; margin-bottom: 7px; }
    .model-benchmark-body .sly-markdown-widget .markdown-body p { margin-bottom: 12px; }
    .model-benchmark-body .el-collapse { margin: 15px 0; box-shadow: 0 1px 2px rgba(0, 0, 0, 0.1); border-radius: 7px; width: fit-content; }
    .model-benchmark-body .el-collapse .el-collapse-item__header { background: transparent; padding-right: 15px; }
    .model-benchmark-body .model-info-block { display: flex; gap: 10px; align-items: center; margin: 0 0 15px; color: #778592; font-size: 13px; }
    .model-benchmark-body .model-info-block > div { display: flex; gap: 4px; align-items: center; }
    /* , .model-benchmark-body .sly-markdown-widget .markdown-body>*:last-child */
    .model-benchmark-body .sly-iw-notification-box .notification-box.notification-box-info { width: fit-content; }
    .model-benchmark-body h1 { font-size: 20px; font-weight: bold; margin-bottom: 5px; }
    .model-benchmark-body .overview-info-block { background: #f4f7fb; width: fit-content; border-radius: 12px; padding: 16px; margin-bottom: 20px; }
    .model-benchmark-body .overview-info-block ul { list-style: none; padding: 0; }
    .model-benchmark-body .overview-info-block ul p { padding: 0; }
    .model-benchmark-body .sly-sidebar-widget .main-wrapper .sidebar-panel { top: 10px; }

    </sly-style>

    <sly-iw-sidebar
        :options="{ height: 'calc(100vh - 130px)', clearMainPanelPaddings: true, leftSided: false,  disableResize: true, sidebarWidth: 300 }"
    >
        <div slot="sidebar">"""

    for vis in metric_visualizations:
        template_str += vis.template_sidebar_str

    template_str += """\n        </div>

        <div style="padding-right: 35px;">"""

    for vis in metric_visualizations:
        template_str += """\n                <div style="margin-top: 20px;">"""
        template_str += vis.template_main_str
        template_str += """\n                </div>"""

    template_str += "\n        </div>\n    </sly-iw-sidebar>"

    template_str += """\n
        <sly-iw-gallery
            ref='modal_general'
            iw-widget-id='modal_general'
            :options="{'isModalWindow': true}"
            :actions="{
                'init': {
                'dataSource': '/data/modal_general.json',
                },
            }"
            :command="command"
            :data="data"
        /> \n
        <sly-iw-gallery
            ref='modal_general_diff'
            iw-widget-id='modal_general_diff'
            :options="{'isModalWindow': true}"
            :actions="{
                'init': {
                'dataSource': '/data/modal_general_diff.json',
                },
                'chart-click': {
                    'dataSource': '/data/gallery_explorer_grid_diff_data.json',
                    'galleryId': 'modal_general',
                        'limit': 3,
                },
            }"
            :command="command"
            :data="data"
        > \n
            <span slot="image-left-header">
                <i class="zmdi zmdi-collection-image"></i> Compare with GT
            </span>
        </sly-iw-gallery>
        </div>"""

    return template_str


template_markdown_str = """
            <div style="margin-top: 10px;">
                <sly-iw-markdown
                id="{{ widget_id }}"
                class="markdown-no-border {{ 'overview-info-block' if is_overview else '' }}"
                iw-widget-id="{{ widget_id }}"
                :actions="{
                    'init': {
                    'dataSource': '{{ data_source }}',
                    },
                }"
                :command="{{ command }}"
                :data="{{ data }}"
                />
            </div>
"""

template_chart_str = """
            <div style="margin-top: 20px; margin-bottom: 20px;">
                <sly-iw-chart
                iw-widget-id="{{ widget_id }}"{% if switchable %}
                v-show="state.{{ radio_group }} === '{{ switch_key }}'"
                {% endif %}:actions="{
                    'init': {
                    'dataSource': '{{ init_data_source }}',
                    },{% if chart_click_data_source %}
                    'chart-click': {
                    'dataSource': '{{ chart_click_data_source }}',{% if cls_name in ['outcome_counts'] %}
                    'getKey': (payload) => payload.points[0].data.name,{% endif %}{% if cls_name in ['frequently_confused', 'recall', 'precision', 'recall_vs_precision'] %}
                    'getKey': (payload) => payload.points[0].label,{% endif %}{% if cls_name in ['pr_curve_by_class'] %}
                    'getKey': (payload) => payload.points[0].data.legendgroup,{% endif %}{% if cls_name in ['per_class_avg_precision'] %}
                    'getKey': (payload) => payload.points[0].theta,{% endif %}{% if cls_name in ['per_class_outcome_counts'] %}
                    'getKey': (payload) => `${payload.points[0].label}${'-'}${payload.points[0].data.name}`,{% endif %}{% if cls_name in ['confusion_matrix', 'per_class_outcome_counts'] %}
                    'keySeparator': '{{ key_separator }}',{% endif %}
                    'galleryId': 'modal_general',
                    'limit': 9
                    },{% endif %}
                }"
                :command="{{ command }}"
                :data="{{ data }}"
                />
            </div>


"""

template_radiogroup_str = """<el-radio v-model="state.{{ radio_group }}" label="{{ switch_key }}" style="margin-top: 10px;">{{ switch_key }}</el-radio>"""


template_gallery_str = """<sly-iw-gallery
            iw-widget-id="{{ widget_id }}"
            {% if is_table_gallery %}
            ref='{{ widget_id }}'
            {% endif %}
            :actions="{
                'init': {
                'dataSource': '{{ init_data_source }}',
                },
                {% if gallery_diff_data_source %}
                'chart-click': {
                    'dataSource': '{{ gallery_diff_data_source }}',
                    'getKey':(payload)=>payload['annotation']['image_id'],
                    'galleryId': 'modal_general',
                        'limit': 3,
                },
                {% endif %}
            }"
            :command="{{ command }}"
            :data="{{ data }}"
            >
                {% if gallery_diff_data_source %}
                    <span slot="image-left-header">
                        <i class="zmdi zmdi-collection-image"></i> Compare with GT
                    </span>
                {% endif %}
            </sly-iw-gallery>

              {% if gallery_click_data_source %}
              <div style="display: flex; justify-content: center; margin-top:10px;" >                
                <el-button iw-widget-id="btn-1" type="primary" @click="command({
                  method: 'update-gallery',
                  payload: {
                    data: {
                      'key': 'explore',
                      'limit': 9,
                      'dataSource': '{{ gallery_click_data_source }}',
                    },
                    'galleryId': 'modal_general_diff',
                  },
                  internalCommand: true
                })">Explore all predictions</el-button>
                </div> {% endif %}
"""


template_table_str = """
            <div style="margin-top: 20px; margin-bottom: 30px; {{ width }}">
                <sly-iw-table
                    iw-widget-id="{{ widget_id }}"
                    {% if clickable %}
                        style="cursor: pointer;"
                    {% endif %}
                    :options="{
                        isRowClickable: '{{ clickable }}' === 'True',
                        fixColumns: {{ fixColumns }}, 
                        showHeaderControls: '{{ showHeaderControls }}' === 'True',
                    }"
                    :actions="{
                    'init': {
                        'dataSource': '{{ init_data_source }}',
                    },
                        {% if clickable %}
                    'chart-click': {
                        'dataSource': '{{ table_click_data }}',
                        'galleryId': '{{ table_gallery_id }}',
                        'getKey':(payload)=>payload.row[0],
                    },
                        {% endif %}
                    }"
                :command="{{ command }}"
                :data="{{ data }}"
                >
                    <span
                    slot="custom-cell-content"
                    slot-scope="{ row, column, cellValue }"
                    >
                    <div
                        v-if="column === ' ' && '{{ mainColumn }}' === 'Batch size'"
                        class="fflex"
                    >
                        <b>Batch size {{ '{{ cellValue }}' }}</b>
                    </div>
                    <div
                        v-if="column === '{{ mainColumn }}'"
                        class="fflex"
                    >
                        <b>{{ '{{ cellValue }}' }}</b>
                    </div>
                    </span>
                </sly-iw-table>
            </div>
            """

template_notification_str = """
            <div style="margin-top: 20px; margin-bottom: 20px;">
                <sly-iw-notification              
                iw-widget-id="{{ widget_id }}"
                :data="{{ data }}"
                >
                <span slot="title">
                    {{ title }}
                </span>

                <span slot="description">
                    {{ description }}
                </span>
                </sly-iw-notification>
            </div>"""



import mimetypes
from pathlib import Path

import magic
import numpy as np
from PIL import Image
from typing import Union, List

from supervisely import Rectangle, Label, logger
from supervisely.imaging.image import read, write
from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    silent_remove,
)

EXT_TO_CONVERT = [".heic", ".avif", ".heif", ".jfif"]


def validate_image(path: str) -> tuple:
    """Validate image by mimetypes, ext, and remove alpha channel."""

    try:
        ext = get_file_ext(path)
        if ext.lower() == ".nrrd":
            return path
        if ext.lower() in EXT_TO_CONVERT:
            path = convert_to_jpg(path)
            return path
        if ext.lower() != ".mpo":
            name = get_file_name_with_ext(path)
            new_name = validate_mimetypes(name, path)
            if new_name != name:
                img = read(path, remove_alpha_channel=False)
                silent_remove(path)
                path = Path(path).with_name(new_name).as_posix()
                write(path, img, remove_alpha_channel=False)
        return path
    except Exception as e:
        logger.warning(f"Skip image: {repr(e)}", extra={"file_path": path})


def validate_mimetypes(name: str, path: str) -> list:
    """Validate mimetypes for images."""

    mimetypes.add_type("image/webp", ".webp")  # to extend types_map
    mimetypes.add_type("image/jpeg", ".jfif")  # to extend types_map

    mime = magic.Magic(mime=True)
    mimetype = mime.from_file(path)
    file_ext = get_file_ext(path).lower()
    if file_ext in mimetypes.guess_all_extensions(mimetype):
        return name

    new_img_ext = mimetypes.guess_extension(mimetype)
    if new_img_ext == ".bin" or new_img_ext is None:
        new_img_ext = ".jpeg"
    new_img_name = f"{get_file_name(name)}{new_img_ext}"
    logger.info(
        f"Image {name} with mimetype {mimetype} will be converted to {new_img_ext}"
    )

    return new_img_name

def convert_to_jpg(path) -> tuple:
    """Convert image to jpg."""

    # * do not remove folllowing imports, it is used to register avif/heic formats
    import pillow_avif  # pylint: disable=import-error
    from pillow_heif import register_heif_opener  # pylint: disable=import-error

    register_heif_opener()

    new_path = Path(path).with_suffix(".jpeg").as_posix()
    with Image.open(path) as image:
        image.convert("RGB").save(new_path)
    silent_remove(path)
    return new_path

def read_tiff_image(path: str) -> Union[np.ndarray, None]:
    """
    Read tiff image.
    Method will transpose image if it has shape (C, H, W) to (H, W, C).
    """

    import tifffile

    logger.debug(f"Found tiff file: {path}.")
    image = tifffile.imread(path)
    name = get_file_name_with_ext(path)
    if image is not None:
        tiff_shape = image.shape
        if image.ndim == 3:
            if tiff_shape[0] < tiff_shape[1] and tiff_shape[0] < tiff_shape[2]:
                image = image.transpose(1, 2, 0)
                logger.warning(
                    f"{name}: transposed shape from {tiff_shape} to {image.shape}"
                )

    return image


def validate_image_bounds(labels: List[Label], img_rect: Rectangle) -> List[Label]:
    """
    Check if labels are localed inside the image canvas, print a warning and skip them if not.
    """
    new_labels = [label for label in labels if img_rect.contains(label.geometry.to_bbox())]
    if new_labels != labels:
        logger.warning(
            f"{len(labels) - len(new_labels)} annotation objects are out of image bounds. Skipping..."
        )
    return new_labels



from typing import Dict

from supervisely.io.fs import ensure_base_path
from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget


class MarkdownWidget(BaseWidget):
    def __init__(
        self,
        name: str,
        title: str,
        text: str = None,
    ) -> None:
        super().__init__(name, title)
        self.text = text
        self.data_source = f"/data/{self.name}_{self.id}.md"
        self.is_info_block = False
        self.width_fit_content = False

    def save_data(self, basepath: str) -> None:
        # init data
        basepath = basepath.rstrip("/")
        ensure_base_path(basepath + self.data_source)

        with open(basepath + self.data_source, "w") as f:
            f.write(self.text)

    def get_state(self) -> Dict:
        return {}

    def to_html(self) -> str:
        style_class = "markdown-no-border"
        if self.is_info_block:
            style_class += " overview-info-block"
        if self.width_fit_content:
            style_class += " markdown-fit-content"

        return f"""
            <div style="margin-top: 10px;">
                <sly-iw-markdown
                id="{ self.id }"
                class="{ style_class }"
                iw-widget-id="{ self.id }"
                :actions="{{
                    'init': {{
                    'dataSource': '{ self.data_source }',
                    }},
                }}"
                :command="command"
                :data="data"
                />
            </div>
        """



from __future__ import annotations

from typing import TYPE_CHECKING, List

from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.visualization.vis_metric_base import MetricVis
from supervisely.nn.benchmark.visualization.vis_widgets import Widget

if TYPE_CHECKING:
    from supervisely.nn.benchmark.visualization.visualizer import Visualizer


class OverallErrorAnalysis(MetricVis):

    def __init__(self, loader: Visualizer) -> None:
        super().__init__(loader)
        self.cv_tasks: List[CVTask] = [CVTask.SEMANTIC_SEGMENTATION.value]

    def get_figure(self, widget: Widget.Chart):  # -> Optional[go.Figure]:
        import plotly.graph_objects as go  # pylint: disable=import-error
        from plotly.subplots import make_subplots  # pylint: disable=import-error

        fig = make_subplots(
            rows=1,
            cols=3,
            subplot_titles=(
                "Basic segmentation metrics",
                "Intersection & Error over Union",
                "Renormalized Error over Union",
            ),
            specs=[[{"type": "polar"}, {"type": "domain"}, {"type": "xy"}]],
        )

        # first subplot
        categories = [
            "mPixel accuracy",
            "mPrecision",
            "mRecall",
            "mF1-score",
            "mIoU",
            "mBoundaryIoU",
            "mPixel accuracy",
        ]
        values = [64, 60.4, 52, 51.4, 37.9, 20.5, 64]
        trace_1 = go.Scatterpolar(
            mode="lines+text",
            r=values,
            theta=categories,
            fill="toself",
            fillcolor="cornflowerblue",
            line_color="blue",
            opacity=0.6,
            text=[64, 60.4, 52, 51.4, 37.9, 20.5, 64],
            textposition=[
                "bottom right",
                "top center",
                "top center",
                "middle left",
                "bottom center",
                "bottom right",
                "bottom right",
            ],
            textfont=dict(color="blue"),
        )
        fig.add_trace(trace_1, row=1, col=1)

        # second subplot
        labels = ["mIoU", "mBoundaryEoU", "mExtentEoU", "mSegmentEoU"]
        values = [37.9, 13.1, 25.8, 23.2]
        trace_2 = go.Pie(
            labels=labels,
            values=values,
            hole=0.5,
            textposition="outside",
            textinfo="percent+label",
            marker=dict(colors=["cornflowerblue", "moccasin", "lightgreen", "orangered"]),
        )
        fig.add_trace(trace_2, row=1, col=2)

        # third subplot
        labels = ["boundary", "extent", "segment"]
        values = [28.9, 37.6, 23.2]
        trace_3 = go.Bar(
            x=labels,
            y=values,
            orientation="v",
            text=values,
            width=[0.5, 0.5, 0.5],
            textposition="outside",
            marker_color=["moccasin", "lightgreen", "orangered"],
        )
        fig.add_trace(trace_3, row=1, col=3)

        fig.update_layout(
            height=400,
            width=1200,
            polar=dict(
                radialaxis=dict(visible=True, showline=False, showticklabels=False, range=[0, 100])
            ),
            showlegend=False,
            plot_bgcolor="rgba(0, 0, 0, 0)",
            yaxis=dict(showticklabels=False),
            yaxis_range=[0, int(max(values)) + 4],
        )
        fig.layout.annotations[0].update(y=1.2)
        fig.layout.annotations[1].update(y=1.2)
        fig.layout.annotations[2].update(y=1.2)

        return fig



# coding: utf-8

from __future__ import annotations

import uuid
from typing import Dict, Optional, Tuple
from uuid import UUID

from supervisely._utils import take_with_default
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely.api.module_api import ApiField
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    LABELER_LOGIN,
    TRACK_ID,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.rectangle import Rectangle
from supervisely.video_annotation.constants import ID, KEY, OBJECT_ID, OBJECT_KEY
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_object import VideoObject
from supervisely.video_annotation.video_object_collection import VideoObjectCollection


class OutOfImageBoundsException(Exception):
    pass


class VideoFigure:
    """
    VideoFigure object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`. :class:`VideoFigure<VideoFigure>` object is immutable.

    :param video_object: VideoObject object.
    :type video_object: VideoObject
    :param geometry: Label :class:`geometry<supervisely.geometry.geometry.Geometry>`.
    :type geometry: Geometry
    :param frame_index: Index of Frame to which VideoFigure belongs.
    :type frame_index: int
    :param key_id_map: KeyIdMap object.
    :type key_id_map: KeyIdMap, optional
    :param class_id: ID of :class:`VideoObject<VideoObject>` to which VideoFigure belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created VideoFigure.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when VideoFigure was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when VideoFigure was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :param track_id: ID of the track to which VideoFigure belongs.
    :type track_id: str, optional
    :param smart_tool_input: Smart Tool parameters that were used for labeling.
    :type smart_tool_input: dict, optional
    :param priority: Priority of the figure (position of the figure relative to other overlapping or underlying figures).
    :type priority: int, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        obj_class_car = sly.ObjClass('car', sly.Rectangle)
        video_obj_car = sly.VideoObject(obj_class_car)
        fr_index = 7
        geometry = sly.Rectangle(0, 0, 100, 100)
        video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
        video_figure_car_json = video_figure_car.to_json()
        print(video_figure_car_json)
        # Output: {
        #     "key": "5e8afd2e26a54ab18154b355fa9665f8",
        #     "objectKey": "5860b7a5519b4de7b3d9c1720a40b38a",
        #     "geometryType": "rectangle",
        #     "geometry": {
        #         "points": {
        #             "exterior": [
        #                 [
        #                     0,
        #                     0
        #                 ],
        #                 [
        #                     100,
        #                     100
        #                 ]
        #             ],
        #             "interior": []
        #         }
        #     }
        # }
    """

    def __init__(
        self,
        video_object: VideoObject,
        geometry: Geometry,
        frame_index: int,
        key: Optional[UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        track_id: Optional[str] = None,
        smart_tool_input: Optional[Dict] = None,
        priority: Optional[int] = None,
    ):
        self._video_object = video_object
        self._set_geometry_inplace(geometry)
        self._frame_index = frame_index
        self._key = take_with_default(key, uuid.uuid4())
        self.class_id = class_id
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at
        self.track_id = track_id
        self._smart_tool_input = smart_tool_input
        self._priority = priority

    def _add_creation_info(self, d):
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at

    def _set_geometry_inplace(self, geometry: Geometry) -> None:
        """
        Checks the given geometry for correctness. Raise error if given geometry type != geometry type of VideoObject class
        :param geometry: Geometry class object (Point, Rectangle etc)
        """
        self._geometry = geometry
        self._validate_geometry_type()
        self._validate_geometry()

    @property
    def video_object(self) -> VideoObject:
        """
        VideoObject of current VideoFigure.

        :return: VideoObject object
        :rtype: :class:`VideoObject<VideoObject>`
        :Usage example:

         .. code-block:: python

            video_obj_car = video_figure_car.video_object
            print(video_obj_car.to_json())
            # Output: {
            #     "key": "d573c6f081544e3da20022d932b259c1",
            #     "classTitle": "car",
            #     "tags": []
            # }
        """
        return self._video_object

    @property
    def parent_object(self) -> VideoObject:
        """
        VideoObject of current VideoFigure.

        :return: VideoObject object
        :rtype: :class:`VideoObject<VideoObject>`
        :Usage example:

         .. code-block:: python

            video_obj_car = video_figure_car.parent_object
            print(video_obj_car.to_json())
            # Output: {
            #     "key": "d573c6f081544e3da20022d932b259c1",
            #     "classTitle": "car",
            #     "tags": []
            # }
        """
        return self._video_object

    @property
    def geometry(self) -> Geometry:
        """
        Geometry of the current VideoFigure.

        :return: Geometry object
        :rtype: :class:`Geometry<supervisely.geometry>`
        :Usage example:

         .. code-block:: python

            geometry = video_figure_car.geometry
            print(geometry.to_json())
            # Output: {
            #     "points": {
            #         "exterior": [
            #             [
            #                 0,
            #                 0
            #             ],
            #             [
            #                 100,
            #                 100
            #             ]
            #         ],
            #         "interior": []
            #     }
            # }
        """
        return self._geometry

    @property
    def frame_index(self) -> int:
        """
        Frame index of the current VideoFigure.

        :return: Index of Frame to which VideoFigure belongs
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            fr_index = video_figure_car.frame_index
            print(fr_index) # 7
        """
        return self._frame_index

    @property
    def smart_tool_input(self):
        """
        Smart Tool parameters that were used for labeling.

        Example:

            {
                'crop': [[85.69912274538524, 323.07711452375236], [1108.5635719011857, 1543.1199742240174]],
                'visible': True,
                'negative': [],
                'positive': [[597, 933], [474.5072466934964, 1381.6437133813354]]
            }
        """
        return self._smart_tool_input

    @smart_tool_input.setter
    def smart_tool_input(self, smtool_input: Dict):
        smtool_input_keys = ["crop", "visible", "negative", "positive"]
        for k in smtool_input_keys:
            if k not in smtool_input:
                raise ValueError(f"Smart tool input has to contain key '{k}'")
        self._smart_tool_input = smtool_input

    @property
    def priority(self):
        """
        Priority of the figure (position of the figure relative to other overlapping or underlying figures).
        """
        return self._priority

    @priority.setter
    def priority(self, priority: int):
        self._priority = priority

    def key(self) -> UUID:
        """
        Figure key.

        :return: Figure key.
        :rtype: UUID
        :Usage example:

         .. code-block:: python

            key = video_figure_car.key
            print(key) # 158e6cf4f4ac4c639fc6994aad127c16
        """

        return self._key

    def _validate_geometry(self):
        """
        Checks geometry of VideoFigure class object for correctness
        """
        self._geometry.validate(
            self.parent_object.obj_class.geometry_type.geometry_name(),
            self.parent_object.obj_class.geometry_config,
        )

    def _validate_geometry_type(self):
        """
        Raise error if given geometry type != geometry type of VideoObject class
        """
        if self.parent_object.obj_class.geometry_type != AnyGeometry:
            if type(self._geometry) is not self.parent_object.obj_class.geometry_type:
                raise RuntimeError(
                    "Input geometry type {!r} != geometry type of ObjClass {}".format(
                        type(self._geometry), self.parent_object.obj_class.geometry_type
                    )
                )

    def to_json(
        self, key_id_map: Optional[KeyIdMap] = None, save_meta: Optional[bool] = False
    ) -> Dict:
        """
        Convert the VideoFigure to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :param save_meta: Save frame index or not.
        :type save_meta: bool, optional
        :return: Json format as a dict
        :rtype: :class:`dict`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
            video_figure_json = video_figure_car.to_json(save_meta=True)
            print(video_figure_json)
            # Output: {
            #     "key": "591d0511ba28462c8cd657691743359c",
            #     "objectKey": "e061bc50bd464c23a008b712d195570a",
            #     "geometryType": "rectangle",
            #     "geometry": {
            #         "points": {
            #             "exterior": [
            #                 [
            #                     0,
            #                     0
            #                 ],
            #                 [
            #                     100,
            #                     100
            #                 ]
            #             ],
            #             "interior": []
            #         }
            #     },
            #     "meta": {
            #         "frame": 7
            #     }
            # }
        """
        data_json = {
            KEY: self.key().hex,
            OBJECT_KEY: self.parent_object.key().hex,
            ApiField.GEOMETRY_TYPE: self.geometry.geometry_name(),
            ApiField.GEOMETRY: self.geometry.to_json(),
        }

        if key_id_map is not None:
            item_id = key_id_map.get_figure_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

            object_id = key_id_map.get_object_id(self.parent_object.key())
            if object_id is not None:
                data_json[OBJECT_ID] = object_id
        if save_meta is True:
            data_json[ApiField.META] = self.get_meta()

        if self._smart_tool_input is not None:
            data_json[ApiField.SMART_TOOL_INPUT] = self._smart_tool_input

        if self._priority is not None:
            data_json[ApiField.PRIORITY] = self._priority

        self._add_creation_info(data_json)
        return data_json

    def get_meta(self) -> Dict[str, int]:
        """
        Get metadata for the video figure.

        :return: Dictionary with metadata for the video figure.
        :rtype: :py:class:`Dict[str, int]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)

            print(video_figure_car.get_meta()) # {'frame': 7}
        """

        return {ApiField.FRAME: self.frame_index}

    @classmethod
    def from_json(
        cls,
        data: Dict,
        objects: VideoObjectCollection,
        frame_index: int,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> VideoFigure:
        """
        Convert a json dict to VideoFigure. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: :class:`dict`
        :param objects: VideoObjectCollection object.
        :type objects: VideoObjectCollection
        :param frame_index: Index of Frame to which VideoFigure belongs.
        :type frame_index: int
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :raises: :class:`RuntimeError`, if video object ID and video object key are None, if video object key and key_id_map are None, if video object with given id not found in key_id_map
        :return: VideoFigure object
        :rtype: :class:`VideoFigure`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Create VideoFigure from json we use data from example to_json(see above)
            new_video_figure = sly.VideoFigure.from_json(video_figure_json, sly.VideoObjectCollection([video_obj_car]), fr_index)
        """
        object_id = data.get(OBJECT_ID, None)
        object_key = None
        if OBJECT_KEY in data:
            object_key = uuid.UUID(data[OBJECT_KEY])

        if object_id is None and object_key is None:
            raise RuntimeError(
                "Figure can not be deserialized from json: object_id or object_key are not found"
            )

        if object_key is None:
            if key_id_map is None:
                raise RuntimeError("Figure can not be deserialized: key_id_map is None")
            object_key = key_id_map.get_object_key(object_id)
            if object_key is None:
                raise RuntimeError(
                    "Object with id={!r} not found in key_id_map".format(object_id)
                )

        object = objects.get(object_key)
        if object is None:
            raise RuntimeError(
                "Figure can not be deserialized: corresponding object {!r} not found in ObjectsCollection".format(
                    object_key.hex
                )
            )

        shape_str = data[ApiField.GEOMETRY_TYPE]
        geometry_json = data[ApiField.GEOMETRY]

        shape = GET_GEOMETRY_FROM_STR(shape_str)
        geometry = shape.from_json(geometry_json)

        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_figure(key, data.get(ID, None))

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        track_id = data.get(TRACK_ID, None)
        smart_tool_input = data.get(ApiField.SMART_TOOL_INPUT, None)
        priority = data.get(ApiField.PRIORITY, None)

        return cls(
            object,
            geometry,
            frame_index,
            key,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
            track_id=track_id,
            smart_tool_input=smart_tool_input,
            priority=priority,
        )

    def clone(
        self,
        video_object: Optional[VideoObject] = None,
        geometry: Optional[Geometry] = None,
        frame_index: Optional[int] = None,
        key: Optional[UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        track_id: Optional[str] = None,
        smart_tool_input: Optional[Dict] = None,
        priority: Optional[int] = None,
    ) -> VideoFigure:
        """
        Makes a copy of VideoFigure with new fields, if fields are given, otherwise it will use fields of the original VideoFigure.

        :param video_object: VideoObject object.
        :type video_object: VideoObject, optional
        :param geometry: Label :class:`geometry<supervisely.geometry.geometry.Geometry>`.
        :type geometry: Geometry, optional
        :param frame_index: Index of Frame to which VideoFigure belongs.
        :type frame_index: int, optional
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which VideoFigure belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created VideoFigure.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when VideoFigure was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when VideoFigure was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :param track_id: ID of the track to which VideoFigure belongs.
        :type track_id: str, optional
        :param smart_tool_input: Smart Tool parameters that were used for labeling.
        :type smart_tool_input: dict, optional
        :param priority: Priority of the figure (position of the figure relative to other overlapping or underlying figures).
        :type priority: int, optional
        :return: VideoFigure object
        :rtype: :class:`VideoFigure`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)

            obj_class_bus = sly.ObjClass('bus', sly.Rectangle)
            video_obj_bus = sly.VideoObject(obj_class_bus)
            fr_index_bus = 15
            geometry_bus = sly.Rectangle(0, 0, 500, 600)

            # Remember that VideoFigure object is immutable, and we need to assign new instance of VideoFigure to a new variable
            video_figure_bus = video_figure_car.clone(video_object=video_obj_bus, geometry=geometry_bus, frame_index=fr_index_bus)
            print(video_figure_bus.to_json())
            # Output: {
            #     "key": "c2f501e94f42483ebd202697608e8d26",
            #     "objectKey": "942c79137b4547c59193276317f73897",
            #     "geometryType": "rectangle",
            #     "geometry": {
            #         "points": {
            #             "exterior": [
            #                 [
            #                     0,
            #                     0
            #                 ],
            #                 [
            #                     600,
            #                     500
            #                 ]
            #             ],
            #             "interior": []
            #         }
            #     }
            # }
        """
        return self.__class__(
            video_object=take_with_default(video_object, self.parent_object),
            geometry=take_with_default(geometry, self.geometry),
            frame_index=take_with_default(frame_index, self.frame_index),
            key=take_with_default(key, self._key),
            class_id=take_with_default(class_id, self.class_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
            track_id=take_with_default(track_id, self.track_id),
            smart_tool_input=take_with_default(smart_tool_input, self._smart_tool_input),
            priority=take_with_default(priority, self._priority),
        )

    def validate_bounds(
        self, img_size: Tuple[int, int], _auto_correct: Optional[bool] = False
    ) -> None:
        """
        Checks if given image with given size contains a figure.

        :param img_size: Size of the image (height, width).
        :type img_size: Tuple[int, int]
        :param _auto_correct: Correct the geometry of a shape if it is out of bounds or not.
        :type _auto_correct: bool, optional
        :raises: :class:`OutOfImageBoundsException<supervisely.video_annotation.video_figure.OutOfImageBoundsException>`, if figure is out of image bounds
        :return: None
        :rtype: :class:`NoneType`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)

            im_size = (50, 200)
            video_figure_car.validate_bounds(im_size)
            # raise OutOfImageBoundsException("Figure is out of image bounds")
        """
        canvas_rect = Rectangle.from_size(img_size)
        if canvas_rect.contains(self.geometry.to_bbox()) is False:
            details = {
                "class_name": self.parent_object.obj_class.name,
                "geometry": self.geometry.geometry_name(),
            }
            if type(self) is VideoFigure:
                details["frame_index"] = self.frame_index
            else:
                details["slice_index"] = self.slice_index  # pylint: disable=no-member
                details["plane"] = self.plane_name  # pylint: disable=no-member
            details_str = ", ".join([f"{k}={v}" for k, v in details.items()])
            raise OutOfImageBoundsException(
                f"Figure is out of image bounds. Figure details: {details_str}."
            )

        if _auto_correct is True:
            geometries_after_crop = [
                cropped_geometry for cropped_geometry in self.geometry.crop(canvas_rect)
            ]
            if len(geometries_after_crop) != 1:
                raise OutOfImageBoundsException("Several geometries after crop")
            self._set_geometry_inplace(geometries_after_crop[0])
